#!/usr/bin/env python
# -*- coding:utf-8 -*- #
#
#   Copyright (c) 2010 Mickaël RAYBAUD
#
#   This program is free software: you can redistribute it and/or modify
#   it under the terms of the GNU General Public License as published by
#   the Free Software Foundation, either version 3 of the License, or
#   (at your option) any later version.
#
#   This program is distributed in the hope that it will be useful,
#   but WITHOUT ANY WARRANTY; without even the implied warranty of
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#   GNU General Public License for more details.
#
#   You should have received a copy of the GNU General Public License
#   along with this program.  If not, see <http://www.gnu.org/licenses/>.
#

__version__='0.01'
__author__=u'Mickaël RAYBAUD <skami@skami-laptop.dyndns.org>'
__doc__="#TODO: penser à faire la documentation..."
import base64
import BaseHTTPServer, SimpleHTTPServer
from random import randint

def extract_parameters(url):
    q=url.split("?")[-1]
    query={}
    if len(q) > 0:
        for pair in q.split('&'):
            if '=' in pair:
                pair=pair.split("=")
                key,value=pair[0], pair[-1]
                if len(key) > 0 and len(value) > 0 :
                    query[key]=value
    return query

class Widget(object): #TODO mieux gérér les erreurs
    """
    Element HTML de base.
    Doit être sous-classé.
    """

    def __init__(self,parent=None,**args):
        """
        Crée une nouvelle instance
        """
        self._cfn_={
            'rowspan':1,
            'colspan':1,
            'visible':True,
            'width':'100%',
            'height':'100%'
        }
        self._parent_=None
        self._needsUpdate_=True
        if isinstance(parent,Widget):
            self._parent_=parent
        elif parent is None:
            pass
        else:
            raise TypeError("Argument `parent' must be a widget !")
        
        self['id']
        self['class']
        for key, value in args.iteritems():
            self._cfn_[key]=value


    def __setitem__(self,item,value):
        self._needsUpdate_=True
        self._cfn_[item] = value

    def __getitem__(self,item):
        item=item.lower()
        if not self._cfn_.has_key(item):
            if item == 'id':
                self._cfn_[item]='widget-%s' % str( id(self) )
            if item == 'class':
                try:    self._cfn_[item]='Wui%s' % self.__class__.__name__
                except: self._cfn_[item]='WuiWidget'
            if item == 'rowspan':
                self._cfn_[item]=1
            if item == 'colspan':
                self._cfn_[item]=1
            if item == 'width':
                self._cfn_[item]='100%'
            if item == 'height':
                self._cfn_[item]='100%'
        return self._cfn_[item]

    def keys(self):           return self._cfn_.keys()
    def has_key(self, key):   return self._cfn_.has_key(key)
    def values(self):         return self._cfn_.values()
    def iteritems(self):      return self._cfn_.iteritems()

    def configure(self,item=None,value=None):
        if not item:
            return self._cfn_
        else:
            return self.__setitem__(item,value)

    def grid(self,*args):
        if self._parent_ and hasattr(self._parent_,'grid'):
            self._parent_.grid(self,*args)

    def html(self):
        """
        Retourne le code html du widget.

        Cette méthode permets de ne pas générer le code HTML à chaque affichage:
          * Si l'attribut `_needsUpdate_' de l'objet est `True', la méthode `__render__' de l'objet est appellée pour générer le code.
        Le code est ensuite stocké dans l'attribut `__html__' de la classe.
          * Si l'attribut `_needsUpdate_' de l'objet est `False', c'est le contenu de la méthode `__html__' qui sera renvoyé.

        Vous pouvez profiter de ces propriétés pour créer des widgets avec un système de cache:
          * Créez une classe qui hérite de cette classe.
          * Créez des méthodes pour modifier les propritétés (la taille où la couleur, par exemple...) de votre nouveau widget, et faites en sorte que chacune de ses méthodes passe l'attribut `_needsUpdate_' de votre classe à `True
          * Ajoutez une méthode `__render__' qui passe l'attribut `_needsUpdate_' à `False' avant de retourner le code HTML qui représente votre widget.

        De cette facon, le code HTML à afficher ne sera caculé qu'une seule fois... mais il sera recalculé chaque fois qu'une propritété du widget change.

        C'est de cette facon que marche le widget `Label' :
        La méthode `setText' passe l'attribut `_needsUpdate_' à `True' chaque fois qu'elle est appellée, de cette facon, c'est le bon texte qui est toujours affiché, car le code HTML est régénéré c'est nécéssaire.

        Bien entendu, vous n'êtes pas obligé d'utiliser ce système de cache basique: vous pouvez très bien vous contentez de créer une méthode `__render__' qui retourne simplement le code HTML adéquat...
        """
        if self._needsUpdate_ is True:
            if hasattr(self,'__render__'):
                print("WUI: update of %s ..." % self.__repr__() )
                self.__html__ = self.__render__()
        return self.__html__



class Table(Widget):
    """
    Container basique pour placer les widgets «en grille».
    """
    __content={0:['']}
    
    def __init__(self,parent=None,**args):
        Widget.__init__(self,parent,**args)
        #self['border']=0

    def __render__(self):
        html="<table border=\"%s\">\n" % str(self['border'])
        for row in self.__content:
            html+="\t<tr>\n"
            for child in self.__content[row]:
                html+='\t\t<td rowspan="%s" colspan="%s">\n' % (str(child['rowspan']), str(child['colspan']))
                if child['visible']:
                    html+=child.html()
                html+="\n\t\t</td>\n"
            html+="\t</tr>\n"
        html+="</table>\n"
        #self._needsUpdate_=False
        return html
            
    def grid(self,widget,row,col):
        """
        Insère un widget aux coordonnées spécifiées.
        """
        if type(row) is not int or type(col) is not int:
            raise TypeError()

        if not isinstance(widget,Widget):
            raise TypeError()

        for i in range(0,row+1):
            if not self.__content.has_key(i):
                self.__content[i] = ['']

        for i in range(0,col+1):
            if i+1 > len(self.__content[row]):
                self.__content[row].append('')

        self.__content[row][col]=widget
            

class Label(Widget):
    """
    Ajoute du texte.
    Si affihez du code html dans ce widget, le code sera insterprété par le navigateur
    """
    def __init__(self,parent=None,text=None, *args):
        """
        Construit une nouvelle instance.
        Prends en argument le texte à afficher...
        """
        Widget.__init__(self,parent,*args)
        self['text'] = text or 'Label-%d'%id(self) 

    def setText(self, text):
        """
        Change le texte affiché
        """
        self['text']=str(text)

    def getText(self):
        """
        Retourne le texte affiché
        """
        return self['text']

    def __render__(self):
        html=str(self['text'])
        self._needsUpdate_=False
        return html

class Button(Label):
    """
    Affiche un bouton pouvant être cliqué.
    Les clics génèrent des évènements qui peuvent être être captés via une instance `WebUI'.
    """

    def __init__(self,parent=None,text=None, url=None, *args):
        Label.__init__(self,parent, text or 'Button-%d'%id(self), *args)
        self['url']=url

    def setUrl(self,url=None):
        self['url']=url

    def getUrl(self):
        return self['url']


    def __render__(self): #TODO gestion des évènements
        url=self['url'] or '/event?widget=%s&event=%s&data=None' % (self['id'], 'clicked')
        html='<a href="%s"><button>%s</button></a>' % (url,self['text'])
        self._needsUpdate_=False
        return html
        
class Image(Widget):
    """
    Affiche une image.
    """
    data=''
    mimetype='image/png'
    tooltip='Image'

    def __init__(self, parent, file='None', b64data=None, mimetype='image/png', *args):
        """
        Construit une nouvelle image.
        Prends en argument le fichier à afficher (encodé en Base64) et son type MIME.
        """
        Widget.__init__(self,parent,*args)
        if b64data is not None: self.setData(b64data)
        if file is not None: self.setDataFromFile(file)

        self.mimetype=mimetype

    def setData(self, b64data):
        self.data=b64data
        self._needsUpdate_ = True

    def setTooltip(self, text):
        self.tooltip=text

    def getTooltip(self):
        return self.tooltip

    def setDataFromFile(self, file):
        data=''
        f=open(file,'rb')
        buffer=f.readline()
        while buffer:
            data+=buffer
            buffer=f.readline()
        self.setData( base64.encodestring(data).replace('\n','') )
        del(data)
        del(buffer)

    def getData(self):
        return self.data

    def __render__(self):
        html='<img src="data:%s;base64,%s" alt="%s" title="%s" />' % (self.mimetype, self.data, self.tooltip, self.tooltip)
        self._needsUpdate_ = False
        return html

class WUIFactory(object):
    """
    Instancie des objets WUI en leur assignant un widget enfant.
    Ne doit pas être utilisé directement.
    """
    def __init__(self,child):
        self.child=child

    def __call__(self, request, client_address, server):
        w=WUI(request, client_address, server, self.child)
        return w


class WUI(BaseHTTPServer.BaseHTTPRequestHandler):
    """
    Propulse des widgets WUI est gère les evènements.
    """
    child=None
    server_version='WUI/%s'%__version__


    def __init__(self, request, client_address, server, child):
        self.child=child
        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, request, client_address, server)


    def do_GET(self):
        self.send_response(200)
        self.send_header("Content-type", "text/html")
        self.end_headers()

        print extract_parameters(self.path)

        self.wfile.write('<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Strict//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd">')
        self.wfile.write('<head>\n\
                         \t<title>WUI</title>\n\
                         \t<meta http-equiv="Content-Type" content="text/html;charset=UTF-8" />\n\
                         \t<meta http-equiv="Pragma" content="no-cache">\n\
                         </head>\n\
                         <body>\n') #TODO rendre personalisable
        # If someone went to "http://something.somewhere.net/foo/bar/",
        # then s.path equals "/foo/bar/".
        self.wfile.write(self.child.html())
        self.wfile.write("</body></html>")
        self.wfile.close()

    
def mainLoop(child,host='localhost',port=8080):
    server = BaseHTTPServer.HTTPServer((host,port),WUIFactory(child))
    server.serve_forever()


if __name__ == '__main__':
    t=Table(border=1)
    title=Label(t,text='<h1>This is WUI.</h1>')
    print title.html()
    title['colspan']=2

    l1=Label(t,text='Label')
    b2=Button(t,text='Button')
    l3=Label(t,text='<iframe src="data:text/plain;text,Plain text"><pre>Plain Text</pre></iframe>')
    l3['colspan']=2

    title.grid(0,0)
    l1.grid(1,0)
    b2.grid(1,1)
    l3.grid(2,0)

    mainLoop(t)
